"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getDiscriminatorConditionFragment = void 0;
const nodes_1 = require("../../../nodes");
const shared_1 = require("../../../shared");
const visitors_1 = require("../../../visitors");
const common_1 = require("./common");
/**
 * ```
 * if (data.length === 72) {
 *   return splTokenAccounts.TOKEN;
 * }
 *
 * if (memcmp(data, getU32Encoder().encode(42), offset)) {
 *   return splTokenAccounts.TOKEN;
 * }
 *
 * if (memcmp(data, new Uint8Array([1, 2, 3]), offset)) {
 *   return splTokenAccounts.TOKEN;
 * }
 * ```
 */
function getDiscriminatorConditionFragment(scope) {
    return (0, common_1.mergeFragments)(scope.discriminators.flatMap((discriminator) => {
        if ((0, nodes_1.isNode)(discriminator, 'sizeDiscriminatorNode')) {
            return [getSizeConditionFragment(discriminator, scope.dataName)];
        }
        if ((0, nodes_1.isNode)(discriminator, 'byteDiscriminatorNode')) {
            return [getByteConditionFragment(discriminator, scope.dataName)];
        }
        if ((0, nodes_1.isNode)(discriminator, 'fieldDiscriminatorNode')) {
            return [getFieldConditionFragment(discriminator, scope)];
        }
        return [];
    }), (r) => r.join(' && ')).mapRender((r) => `if (${r}) { ${scope.ifTrue}; }`);
}
exports.getDiscriminatorConditionFragment = getDiscriminatorConditionFragment;
function getSizeConditionFragment(discriminator, dataName) {
    return (0, common_1.fragment)(`${dataName}.length === ${discriminator.size}`);
}
function getByteConditionFragment(discriminator, dataName) {
    const bytes = discriminator.bytes.join(', ');
    return (0, common_1.fragment)(`memcmp(${dataName}, new Uint8Array([${bytes}]), ${discriminator.offset})`).addImports('shared', 'memcmp');
}
function getFieldConditionFragment(discriminator, scope) {
    const field = scope.struct.fields.find((f) => f.name === discriminator.name);
    if (!field || !field.defaultValue) {
        throw new shared_1.InvalidKinobiTreeError(`Field discriminator "${discriminator.name}" does not have a matching argument with default value.`);
    }
    // This handles the case where a field uses an u8 array to represent its discriminator.
    // In this case, we can simplify the generated code by delegating to a byteDiscriminatorNode.
    if ((0, nodes_1.isNode)(field.type, 'arrayTypeNode') &&
        (0, nodes_1.isNode)(field.type.item, 'numberTypeNode') &&
        field.type.item.format === 'u8' &&
        (0, nodes_1.isNode)(field.type.size, 'fixedSizeNode') &&
        (0, nodes_1.isNode)(field.defaultValue, 'arrayValueNode') &&
        field.defaultValue.items.every((0, nodes_1.isNodeFilter)('numberValueNode'))) {
        return getByteConditionFragment((0, nodes_1.byteDiscriminatorNode)(field.defaultValue.items.map((node) => node.number), discriminator.offset), scope.dataName);
    }
    return (0, common_1.mergeFragments)([
        (0, visitors_1.visit)(field.type, scope.typeManifestVisitor).encoder,
        (0, visitors_1.visit)(field.defaultValue, scope.valueNodeVisitor),
    ], ([encoderFunction, value]) => `${encoderFunction}.encode(${value})`)
        .mapRender((r) => `memcmp(${scope.dataName}, ${r}, ${discriminator.offset})`)
        .addImports('shared', 'memcmp');
}
//# sourceMappingURL=discriminatorCondition.js.map