"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getTypeManifestVisitor = void 0;
const nodes_1 = require("../../nodes");
const shared_1 = require("../../shared");
const visitors_1 = require("../../visitors");
const ImportMap_1 = require("./ImportMap");
const TypeManifest_1 = require("./TypeManifest");
const fragments_1 = require("./fragments");
function getTypeManifestVisitor(input) {
    var _a;
    const { nameApi, valueNodeVisitor, customAccountData, customInstructionData, } = input;
    let parentName = (_a = input.parentName) !== null && _a !== void 0 ? _a : null;
    return (0, shared_1.pipe)((0, visitors_1.staticVisitor)(() => ({
        isEnum: false,
        strictType: (0, fragments_1.fragment)(''),
        looseType: (0, fragments_1.fragment)(''),
        encoder: (0, fragments_1.fragment)(''),
        decoder: (0, fragments_1.fragment)(''),
    }), [
        ...nodes_1.REGISTERED_TYPE_NODE_KINDS,
        'definedTypeLinkNode',
        'definedTypeNode',
        'accountNode',
        'instructionNode',
    ]), (visitor) => (0, visitors_1.extendVisitor)(visitor, {
        visitAccount(account, { self }) {
            var _a;
            const accountDataName = nameApi.accountDataType(account.name);
            parentName = {
                strict: nameApi.dataType(accountDataName),
                loose: nameApi.dataArgsType(accountDataName),
            };
            const link = (_a = customAccountData.get(account.name)) === null || _a === void 0 ? void 0 : _a.linkNode;
            const manifest = link ? (0, visitors_1.visit)(link, self) : (0, visitors_1.visit)(account.data, self);
            parentName = null;
            return manifest;
        },
        visitInstruction(instruction, { self }) {
            var _a;
            const instructionDataName = nameApi.instructionDataType(instruction.name);
            parentName = {
                strict: nameApi.dataType(instructionDataName),
                loose: nameApi.dataArgsType(instructionDataName),
            };
            const link = (_a = customInstructionData.get(instruction.name)) === null || _a === void 0 ? void 0 : _a.linkNode;
            const struct = (0, nodes_1.structTypeNodeFromInstructionArgumentNodes)(instruction.arguments);
            const manifest = link ? (0, visitors_1.visit)(link, self) : (0, visitors_1.visit)(struct, self);
            parentName = null;
            return manifest;
        },
        visitDefinedType(definedType, { self }) {
            parentName = {
                strict: nameApi.dataType(definedType.name),
                loose: nameApi.dataArgsType(definedType.name),
            };
            const manifest = (0, visitors_1.visit)(definedType.type, self);
            parentName = null;
            return manifest;
        },
        visitArrayType(arrayType, { self }) {
            const childManifest = (0, visitors_1.visit)(arrayType.item, self);
            childManifest.looseType.mapRender((r) => `Array<${r}>`);
            childManifest.strictType.mapRender((r) => `Array<${r}>`);
            const sizeManifest = getArrayLikeSizeOption(arrayType.size, self);
            const encoderOptions = sizeManifest.encoder.render
                ? `, { ${sizeManifest.encoder.render} }`
                : '';
            const decoderOptions = sizeManifest.decoder.render
                ? `, { ${sizeManifest.decoder.render} }`
                : '';
            childManifest.encoder
                .mapRender((r) => `getArrayEncoder(${r + encoderOptions})`)
                .mergeImportsWith(sizeManifest.encoder)
                .addImports('solanaCodecsDataStructures', 'getArrayEncoder');
            childManifest.decoder
                .mapRender((r) => `getArrayDecoder(${r + decoderOptions})`)
                .mergeImportsWith(sizeManifest.decoder)
                .addImports('solanaCodecsDataStructures', 'getArrayDecoder');
            return childManifest;
        },
        visitDefinedTypeLink(node) {
            var _a;
            const strictName = nameApi.dataType(node.name);
            const looseName = nameApi.dataArgsType(node.name);
            const encoderFunction = nameApi.encoderFunction(node.name);
            const decoderFunction = nameApi.decoderFunction(node.name);
            const importFrom = (_a = node.importFrom) !== null && _a !== void 0 ? _a : 'generatedTypes';
            return {
                isEnum: false,
                strictType: (0, fragments_1.fragment)(strictName).addImports(importFrom, strictName),
                looseType: (0, fragments_1.fragment)(looseName).addImports(importFrom, looseName),
                encoder: (0, fragments_1.fragment)(`${encoderFunction}()`).addImports(importFrom, encoderFunction),
                decoder: (0, fragments_1.fragment)(`${decoderFunction}()`).addImports(importFrom, decoderFunction),
            };
        },
        visitEnumType(enumType, { self }) {
            const currentParentName = parentName;
            parentName = null;
            const encoderImports = new ImportMap_1.ImportMap();
            const decoderImports = new ImportMap_1.ImportMap();
            const encoderOptions = [];
            const decoderOptions = [];
            if (enumType.size.format !== 'u8' || enumType.size.endian !== 'le') {
                const sizeManifest = (0, visitors_1.visit)(enumType.size, self);
                encoderImports.mergeWith(sizeManifest.encoder);
                decoderImports.mergeWith(sizeManifest.decoder);
                encoderOptions.push(`size: ${sizeManifest.encoder.render}`);
                decoderOptions.push(`size: ${sizeManifest.decoder.render}`);
            }
            const encoderOptionsAsString = encoderOptions.length > 0
                ? `, { ${encoderOptions.join(', ')} }`
                : '';
            const decoderOptionsAsString = decoderOptions.length > 0
                ? `, { ${decoderOptions.join(', ')} }`
                : '';
            if ((0, nodes_1.isScalarEnum)(enumType)) {
                if (currentParentName === null) {
                    throw new Error('Scalar enums cannot be inlined and must be introduced ' +
                        'via a defined type. Ensure you are not inlining a ' +
                        'defined type that is a scalar enum through a visitor.');
                }
                const variantNames = enumType.variants.map(({ name }) => nameApi.scalarEnumVariant(name));
                return {
                    isEnum: true,
                    strictType: (0, fragments_1.fragment)(`{ ${variantNames.join(', ')} }`),
                    looseType: (0, fragments_1.fragment)(`{ ${variantNames.join(', ')} }`),
                    encoder: (0, fragments_1.fragment)(`getScalarEnumEncoder(${currentParentName.strict + encoderOptionsAsString})`, encoderImports.add('solanaCodecsDataStructures', 'getScalarEnumEncoder')),
                    decoder: (0, fragments_1.fragment)(`getScalarEnumDecoder(${currentParentName.strict + decoderOptionsAsString})`, decoderImports.add('solanaCodecsDataStructures', 'getScalarEnumDecoder')),
                };
            }
            const mergedManifest = (0, TypeManifest_1.mergeManifests)(enumType.variants.map((variant) => (0, visitors_1.visit)(variant, self)), (renders) => renders.join(' | '), (renders) => renders.join(', '));
            mergedManifest.encoder
                .mapRender((r) => `getDataEnumEncoder([${r}]${encoderOptionsAsString})`)
                .mergeImportsWith(encoderImports)
                .addImports('solanaCodecsDataStructures', ['getDataEnumEncoder']);
            mergedManifest.decoder
                .mapRender((r) => `getDataEnumDecoder([${r}]${decoderOptionsAsString})`)
                .mergeImportsWith(decoderImports)
                .addImports('solanaCodecsDataStructures', ['getDataEnumDecoder']);
            return mergedManifest;
        },
        visitEnumEmptyVariantType(enumEmptyVariantType) {
            const discriminator = nameApi.dataEnumDiscriminator(enumEmptyVariantType.name);
            const name = nameApi.dataEnumVariant(enumEmptyVariantType.name);
            const kindAttribute = `${discriminator}: "${name}"`;
            return {
                isEnum: false,
                strictType: (0, fragments_1.fragment)(`{ ${kindAttribute} }`),
                looseType: (0, fragments_1.fragment)(`{ ${kindAttribute} }`),
                encoder: (0, fragments_1.fragment)(`['${name}', getUnitEncoder()]`).addImports('solanaCodecsDataStructures', 'getUnitEncoder'),
                decoder: (0, fragments_1.fragment)(`['${name}', getUnitDecoder()]`).addImports('solanaCodecsDataStructures', 'getUnitDecoder'),
            };
        },
        visitEnumStructVariantType(enumStructVariantType, { self }) {
            const discriminator = nameApi.dataEnumDiscriminator(enumStructVariantType.name);
            const name = nameApi.dataEnumVariant(enumStructVariantType.name);
            const kindAttribute = `${discriminator}: "${name}"`;
            const structManifest = (0, visitors_1.visit)(enumStructVariantType.struct, self);
            structManifest.strictType.mapRender((r) => `{ ${kindAttribute},${r.slice(1, -1)}}`);
            structManifest.looseType.mapRender((r) => `{ ${kindAttribute},${r.slice(1, -1)}}`);
            structManifest.encoder.mapRender((r) => `['${name}', ${r}]`);
            structManifest.decoder.mapRender((r) => `['${name}', ${r}]`);
            return structManifest;
        },
        visitEnumTupleVariantType(enumTupleVariantType, { self }) {
            const discriminator = nameApi.dataEnumDiscriminator(enumTupleVariantType.name);
            const name = nameApi.dataEnumVariant(enumTupleVariantType.name);
            const kindAttribute = `${discriminator}: "${name}"`;
            const struct = (0, nodes_1.structTypeNode)([
                (0, nodes_1.structFieldTypeNode)({
                    name: 'fields',
                    type: enumTupleVariantType.tuple,
                }),
            ]);
            const structManifest = (0, visitors_1.visit)(struct, self);
            structManifest.strictType.mapRender((r) => `{ ${kindAttribute},${r.slice(1, -1)}}`);
            structManifest.looseType.mapRender((r) => `{ ${kindAttribute},${r.slice(1, -1)}}`);
            structManifest.encoder.mapRender((r) => `['${name}', ${r}]`);
            structManifest.decoder.mapRender((r) => `['${name}', ${r}]`);
            return structManifest;
        },
        visitMapType(mapType, { self }) {
            const key = (0, visitors_1.visit)(mapType.key, self);
            const value = (0, visitors_1.visit)(mapType.value, self);
            const mergedManifest = (0, TypeManifest_1.mergeManifests)([key, value], ([k, v]) => `Map<${k}, ${v}>`, ([k, v]) => `${k}, ${v}`);
            const sizeManifest = getArrayLikeSizeOption(mapType.size, self);
            const encoderOptions = sizeManifest.encoder.render
                ? `, { ${sizeManifest.encoder.render} }`
                : '';
            const decoderOptions = sizeManifest.decoder.render
                ? `, { ${sizeManifest.decoder.render} }`
                : '';
            mergedManifest.encoder
                .mapRender((r) => `getMapEncoder(${r}${encoderOptions})`)
                .addImports('solanaCodecsDataStructures', 'getMapEncoder');
            mergedManifest.decoder
                .mapRender((r) => `getMapDecoder(${r}${decoderOptions})`)
                .addImports('solanaCodecsDataStructures', 'getMapDecoder');
            return mergedManifest;
        },
        visitOptionType(optionType, { self }) {
            const childManifest = (0, visitors_1.visit)(optionType.item, self);
            childManifest.strictType
                .mapRender((r) => `Option<${r}>`)
                .addImports('solanaOptions', 'Option');
            childManifest.looseType
                .mapRender((r) => `OptionOrNullable<${r}>`)
                .addImports('solanaOptions', 'OptionOrNullable');
            const encoderOptions = [];
            const decoderOptions = [];
            // Prefix option.
            if (optionType.prefix.format !== 'u8' ||
                optionType.prefix.endian !== 'le') {
                const prefixManifest = (0, visitors_1.visit)(optionType.prefix, self);
                childManifest.encoder.mergeImportsWith(prefixManifest.encoder);
                childManifest.decoder.mergeImportsWith(prefixManifest.decoder);
                encoderOptions.push(`prefix: ${prefixManifest.encoder.render}`);
                decoderOptions.push(`prefix: ${prefixManifest.decoder.render}`);
            }
            // Fixed option.
            if (optionType.fixed) {
                encoderOptions.push(`fixed: true`);
                decoderOptions.push(`fixed: true`);
            }
            const encoderOptionsAsString = encoderOptions.length > 0
                ? `, { ${encoderOptions.join(', ')} }`
                : '';
            const decoderOptionsAsString = decoderOptions.length > 0
                ? `, { ${decoderOptions.join(', ')} }`
                : '';
            childManifest.encoder
                .mapRender((r) => `getOptionEncoder(${r + encoderOptionsAsString})`)
                .addImports('solanaOptions', 'getOptionEncoder');
            childManifest.decoder
                .mapRender((r) => `getOptionDecoder(${r + decoderOptionsAsString})`)
                .addImports('solanaOptions', 'getOptionDecoder');
            return childManifest;
        },
        visitSetType(setType, { self }) {
            const childManifest = (0, visitors_1.visit)(setType.item, self);
            childManifest.strictType.mapRender((r) => `Set<${r}>`);
            childManifest.looseType.mapRender((r) => `Set<${r}>`);
            const sizeManifest = getArrayLikeSizeOption(setType.size, self);
            const encoderOptions = sizeManifest.encoder.render
                ? `, { ${sizeManifest.encoder.render} }`
                : '';
            const decoderOptions = sizeManifest.decoder.render
                ? `, { ${sizeManifest.decoder.render} }`
                : '';
            childManifest.encoder
                .mergeImportsWith(sizeManifest.encoder)
                .mapRender((r) => `getSetEncoder(${r + encoderOptions})`)
                .addImports('solanaCodecsDataStructures', 'getSetEncoder');
            childManifest.decoder
                .mergeImportsWith(sizeManifest.decoder)
                .mapRender((r) => `getSetDecoder(${r + decoderOptions})`)
                .addImports('solanaCodecsDataStructures', 'getSetDecoder');
            return childManifest;
        },
        visitStructType(structType, { self }) {
            // const currentParentName = parentName;
            parentName = null;
            const optionalFields = structType.fields.filter((f) => !!f.defaultValue);
            const mergedManifest = (0, TypeManifest_1.mergeManifests)(structType.fields.map((field) => (0, visitors_1.visit)(field, self)), (renders) => `{ ${renders.join('')} }`, (renders) => `([${renders.join(', ')}])`);
            mergedManifest.encoder
                .mapRender((r) => `getStructEncoder${r}`)
                .addImports('solanaCodecsDataStructures', 'getStructEncoder');
            mergedManifest.decoder
                .mapRender((r) => `getStructDecoder${r}`)
                .addImports('solanaCodecsDataStructures', 'getStructDecoder');
            if (optionalFields.length === 0) {
                return mergedManifest;
            }
            const defaultValues = optionalFields
                .map((f) => {
                const key = (0, shared_1.camelCase)(f.name);
                const defaultValue = f.defaultValue;
                const { render: renderedValue, imports } = (0, visitors_1.visit)(defaultValue, valueNodeVisitor);
                mergedManifest.encoder.mergeImportsWith(imports);
                return f.defaultValueStrategy === 'omitted'
                    ? `${key}: ${renderedValue}`
                    : `${key}: value.${key} ?? ${renderedValue}`;
            })
                .join(', ');
            mergedManifest.encoder
                .mapRender((r) => `mapEncoder(${r}, (value) => ({ ...value, ${defaultValues} }))`)
                .addImports('solanaCodecsCore', 'mapEncoder');
            return mergedManifest;
        },
        visitStructFieldType(structFieldType, { self }) {
            const name = (0, shared_1.camelCase)(structFieldType.name);
            const childManifest = (0, visitors_1.visit)(structFieldType.type, self);
            const docblock = structFieldType.docs.length > 0
                ? `\n${(0, shared_1.jsDocblock)(structFieldType.docs)}`
                : '';
            const originalLooseType = childManifest.looseType.render;
            childManifest.strictType.mapRender((r) => `${docblock}${name}: ${r}; `);
            childManifest.looseType.mapRender((r) => `${docblock}${name}: ${r}; `);
            childManifest.encoder.mapRender((r) => `['${name}', ${r}]`);
            childManifest.decoder.mapRender((r) => `['${name}', ${r}]`);
            // No default value.
            if (!structFieldType.defaultValue) {
                return childManifest;
            }
            // Optional default value.
            if (structFieldType.defaultValueStrategy !== 'omitted') {
                childManifest.looseType.setRender(`${docblock}${name}?: ${originalLooseType}; `);
                return childManifest;
            }
            // Omitted default value.
            childManifest.looseType = (0, fragments_1.fragment)('');
            return childManifest;
        },
        visitTupleType(tupleType, { self }) {
            const items = tupleType.items.map((item) => (0, visitors_1.visit)(item, self));
            const mergedManifest = (0, TypeManifest_1.mergeManifests)(items, (types) => `[${types.join(', ')}]`, (codecs) => `[${codecs.join(', ')}]`);
            mergedManifest.encoder
                .mapRender((render) => `getTupleEncoder(${render})`)
                .addImports('solanaCodecsDataStructures', 'getTupleEncoder');
            mergedManifest.decoder
                .mapRender((render) => `getTupleDecoder(${render})`)
                .addImports('solanaCodecsDataStructures', 'getTupleDecoder');
            return mergedManifest;
        },
        visitBooleanType(booleanType, { self }) {
            const encoderImports = new ImportMap_1.ImportMap().add('solanaCodecsDataStructures', 'getBooleanEncoder');
            const decoderImports = new ImportMap_1.ImportMap().add('solanaCodecsDataStructures', 'getBooleanDecoder');
            let sizeEncoder = '';
            let sizeDecoder = '';
            if (booleanType.size.format !== 'u8' ||
                booleanType.size.endian !== 'le') {
                const size = (0, visitors_1.visit)(booleanType.size, self);
                encoderImports.mergeWith(size.encoder);
                decoderImports.mergeWith(size.decoder);
                sizeEncoder = `{ size: ${size.encoder.render} }`;
                sizeDecoder = `{ size: ${size.decoder.render} }`;
            }
            return {
                isEnum: false,
                strictType: (0, fragments_1.fragment)('boolean'),
                looseType: (0, fragments_1.fragment)('boolean'),
                encoder: (0, fragments_1.fragment)(`getBooleanEncoder(${sizeEncoder})`, encoderImports),
                decoder: (0, fragments_1.fragment)(`getBooleanDecoder(${sizeDecoder})`, decoderImports),
            };
        },
        visitBytesType(bytesType, { self }) {
            const encoderImports = new ImportMap_1.ImportMap().add('solanaCodecsDataStructures', 'getBytesEncoder');
            const decoderImports = new ImportMap_1.ImportMap().add('solanaCodecsDataStructures', 'getBytesDecoder');
            const encoderOptions = [];
            const decoderOptions = [];
            // Size option.
            if ((0, nodes_1.isNode)(bytesType.size, 'prefixedSizeNode')) {
                const prefix = (0, visitors_1.visit)(bytesType.size.prefix, self);
                encoderImports.mergeWith(prefix.encoder);
                decoderImports.mergeWith(prefix.decoder);
                encoderOptions.push(`size: ${prefix.encoder.render}`);
                decoderOptions.push(`size: ${prefix.decoder.render}`);
            }
            else if ((0, nodes_1.isNode)(bytesType.size, 'fixedSizeNode')) {
                encoderOptions.push(`size: ${bytesType.size.size}`);
                decoderOptions.push(`size: ${bytesType.size.size}`);
            }
            const encoderOptionsAsString = encoderOptions.length > 0 ? `{ ${encoderOptions.join(', ')} }` : '';
            const decoderOptionsAsString = decoderOptions.length > 0 ? `{ ${decoderOptions.join(', ')} }` : '';
            return {
                isEnum: false,
                strictType: (0, fragments_1.fragment)('Uint8Array'),
                looseType: (0, fragments_1.fragment)('Uint8Array'),
                encoder: (0, fragments_1.fragment)(`getBytesEncoder(${encoderOptionsAsString})`, encoderImports),
                decoder: (0, fragments_1.fragment)(`getBytesDecoder(${decoderOptionsAsString})`, decoderImports),
            };
        },
        visitNumberType(numberType) {
            const encoderFunction = nameApi.encoderFunction(numberType.format);
            const decoderFunction = nameApi.decoderFunction(numberType.format);
            const isBigNumber = ['u64', 'u128', 'i64', 'i128'].includes(numberType.format);
            const encoderImports = new ImportMap_1.ImportMap().add('solanaCodecsNumbers', encoderFunction);
            const decoderImports = new ImportMap_1.ImportMap().add('solanaCodecsNumbers', decoderFunction);
            let endianness = '';
            if (numberType.endian === 'be') {
                encoderImports.add('solanaCodecsNumbers', 'Endian');
                decoderImports.add('solanaCodecsNumbers', 'Endian');
                endianness = '{ endian: Endian.Big }';
            }
            return {
                isEnum: false,
                strictType: (0, fragments_1.fragment)(isBigNumber ? 'bigint' : 'number'),
                looseType: (0, fragments_1.fragment)(isBigNumber ? 'number | bigint' : 'number'),
                encoder: (0, fragments_1.fragment)(`${encoderFunction}(${endianness})`, encoderImports),
                decoder: (0, fragments_1.fragment)(`${decoderFunction}(${endianness})`, decoderImports),
            };
        },
        visitAmountType(amountType, { self }) {
            return (0, visitors_1.visit)(amountType.number, self);
        },
        visitDateTimeType(dateTimeType, { self }) {
            return (0, visitors_1.visit)(dateTimeType.number, self);
        },
        visitSolAmountType(solAmountType, { self }) {
            return (0, visitors_1.visit)(solAmountType.number, self);
        },
        visitPublicKeyType() {
            const imports = new ImportMap_1.ImportMap().add('solanaAddresses', 'Address');
            return {
                isEnum: false,
                strictType: (0, fragments_1.fragment)('Address', imports),
                looseType: (0, fragments_1.fragment)('Address', imports),
                encoder: (0, fragments_1.fragment)('getAddressEncoder()').addImports('solanaAddresses', 'getAddressEncoder'),
                decoder: (0, fragments_1.fragment)('getAddressDecoder()').addImports('solanaAddresses', 'getAddressDecoder'),
            };
        },
        visitStringType(stringType, { self }) {
            const encoderImports = new ImportMap_1.ImportMap().add('solanaCodecsStrings', 'getStringEncoder');
            const decoderImports = new ImportMap_1.ImportMap().add('solanaCodecsStrings', 'getStringDecoder');
            const encoderOptions = [];
            const decoderOptions = [];
            // Encoding option.
            if (stringType.encoding !== 'utf8') {
                const encoderFunction = nameApi.encoderFunction(stringType.encoding);
                const decoderFunction = nameApi.decoderFunction(stringType.encoding);
                encoderImports.add('solanaCodecsStrings', encoderFunction);
                decoderImports.add('solanaCodecsStrings', decoderFunction);
                encoderOptions.push(`encoding: ${encoderFunction}`);
                decoderOptions.push(`encoding: ${decoderFunction}`);
            }
            // Size option.
            if ((0, nodes_1.isNode)(stringType.size, 'remainderSizeNode')) {
                encoderOptions.push(`size: 'variable'`);
                decoderOptions.push(`size: 'variable'`);
            }
            else if ((0, nodes_1.isNode)(stringType.size, 'fixedSizeNode')) {
                encoderOptions.push(`size: ${stringType.size.size}`);
                decoderOptions.push(`size: ${stringType.size.size}`);
            }
            else if (stringType.size.prefix.format !== 'u32' ||
                stringType.size.prefix.endian !== 'le') {
                const prefix = (0, visitors_1.visit)(stringType.size.prefix, self);
                encoderImports.mergeWith(prefix.encoder.imports);
                decoderImports.mergeWith(prefix.decoder.imports);
                encoderOptions.push(`size: ${prefix.encoder.render}`);
                decoderOptions.push(`size: ${prefix.decoder.render}`);
            }
            const encoderOptionsAsString = encoderOptions.length > 0 ? `{ ${encoderOptions.join(', ')} }` : '';
            const decoderOptionsAsString = decoderOptions.length > 0 ? `{ ${decoderOptions.join(', ')} }` : '';
            return {
                isEnum: false,
                strictType: (0, fragments_1.fragment)('string'),
                looseType: (0, fragments_1.fragment)('string'),
                encoder: (0, fragments_1.fragment)(`getStringEncoder(${encoderOptionsAsString})`, encoderImports),
                decoder: (0, fragments_1.fragment)(`getStringDecoder(${decoderOptionsAsString})`, decoderImports),
            };
        },
    }));
}
exports.getTypeManifestVisitor = getTypeManifestVisitor;
function getArrayLikeSizeOption(size, visitor) {
    if ((0, nodes_1.isNode)(size, 'fixedSizeNode')) {
        return {
            encoder: (0, fragments_1.fragment)(`size: ${size.size}`),
            decoder: (0, fragments_1.fragment)(`size: ${size.size}`),
        };
    }
    if ((0, nodes_1.isNode)(size, 'remainderSizeNode')) {
        return {
            encoder: (0, fragments_1.fragment)(`size: 'remainder'`),
            decoder: (0, fragments_1.fragment)(`size: 'remainder'`),
        };
    }
    if (size.prefix.format === 'u32' && size.prefix.endian === 'le') {
        return { encoder: (0, fragments_1.fragment)(''), decoder: (0, fragments_1.fragment)('') };
    }
    const prefixManifest = (0, visitors_1.visit)(size.prefix, visitor);
    prefixManifest.encoder.mapRender((r) => `size: ${r}`);
    prefixManifest.decoder.mapRender((r) => `size: ${r}`);
    return prefixManifest;
}
//# sourceMappingURL=getTypeManifestVisitor.js.map