"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.renderInstructionDefaults = void 0;
const nodes_1 = require("../../nodes");
const shared_1 = require("../../shared");
const visitors_1 = require("../../visitors");
const JavaScriptContextMap_1 = require("./JavaScriptContextMap");
const JavaScriptImportMap_1 = require("./JavaScriptImportMap");
function renderInstructionDefaults(input, valueNodeVisitor, optionalAccountStrategy, argObject) {
    var _a, _b, _c, _d;
    const imports = new JavaScriptImportMap_1.JavaScriptImportMap();
    const interfaces = new JavaScriptContextMap_1.JavaScriptContextMap();
    if (!input.defaultValue) {
        return { imports, interfaces, render: '' };
    }
    const { defaultValue } = input;
    const render = (renderedValue, isWritable) => {
        const inputName = (0, shared_1.camelCase)(input.name);
        if (input.kind === 'instructionAccountNode' &&
            (0, nodes_1.isNode)(defaultValue, 'resolverValueNode')) {
            return {
                imports,
                interfaces,
                render: `resolvedAccounts.${inputName} = { ...resolvedAccounts.${inputName}, ...${renderedValue} };`,
            };
        }
        if (input.kind === 'instructionAccountNode' && isWritable === undefined) {
            return {
                imports,
                interfaces,
                render: `resolvedAccounts.${inputName}.value = ${renderedValue};`,
            };
        }
        if (input.kind === 'instructionAccountNode') {
            return {
                imports,
                interfaces,
                render: `resolvedAccounts.${inputName}.value = ${renderedValue};\n` +
                    `resolvedAccounts.${inputName}.isWritable = ${isWritable ? 'true' : 'false'}`,
            };
        }
        return {
            imports,
            interfaces,
            render: `${argObject}.${inputName} = ${renderedValue};`,
        };
    };
    switch (defaultValue.kind) {
        case 'accountValueNode':
            const name = (0, shared_1.camelCase)(defaultValue.name);
            if (input.kind === 'instructionAccountNode') {
                imports.add('shared', 'expectSome');
                if (input.resolvedIsSigner && !input.isSigner) {
                    return render(`expectSome(resolvedAccounts.${name}.value).publicKey`);
                }
                return render(`expectSome(resolvedAccounts.${name}.value)`);
            }
            imports.add('shared', 'expectPublicKey');
            return render(`expectPublicKey(resolvedAccounts.${name}.value)`);
        case 'pdaValueNode':
            const pdaFunction = `find${(0, shared_1.pascalCase)(defaultValue.pda.name)}Pda`;
            const pdaImportFrom = (_a = defaultValue.pda.importFrom) !== null && _a !== void 0 ? _a : 'generatedAccounts';
            imports.add(pdaImportFrom, pdaFunction);
            interfaces.add('eddsa');
            const pdaArgs = ['context'];
            const pdaSeeds = defaultValue.seeds.map((seed) => {
                if ((0, nodes_1.isNode)(seed.value, 'accountValueNode')) {
                    imports.add('shared', 'expectPublicKey');
                    return `${seed.name}: expectPublicKey(resolvedAccounts.${(0, shared_1.camelCase)(seed.value.name)}.value)`;
                }
                if ((0, nodes_1.isNode)(seed.value, 'argumentValueNode')) {
                    imports.add('shared', 'expectSome');
                    return `${seed.name}: expectSome(${argObject}.${(0, shared_1.camelCase)(seed.value.name)})`;
                }
                const valueManifest = (0, visitors_1.visit)(seed.value, valueNodeVisitor);
                imports.mergeWith(valueManifest.imports);
                return `${seed.name}: ${valueManifest.render}`;
            });
            if (pdaSeeds.length > 0) {
                pdaArgs.push(`{ ${pdaSeeds.join(', ')} }`);
            }
            return render(`${pdaFunction}(${pdaArgs.join(', ')})`);
        case 'publicKeyValueNode':
            if (!defaultValue.identifier) {
                imports.add('umi', 'publicKey');
                return render(`publicKey('${defaultValue.publicKey}')`);
            }
            interfaces.add('programs');
            return render(`context.programs.getPublicKey('${defaultValue.identifier}', '${defaultValue.publicKey}')`, false);
        case 'programLinkNode':
            const importFrom = (_b = defaultValue.importFrom) !== null && _b !== void 0 ? _b : 'generatedPrograms';
            const functionName = `get${(0, shared_1.pascalCase)(defaultValue.name)}ProgramId`;
            imports.add(importFrom, functionName);
            return render(`${functionName}(context)`, false);
        case 'programIdValueNode':
            if (optionalAccountStrategy === 'programId' &&
                input.kind === 'instructionAccountNode' &&
                input.isOptional) {
                return { imports, interfaces, render: '' };
            }
            return render('programId', false);
        case 'identityValueNode':
            interfaces.add('identity');
            if (input.kind === 'instructionAccountNode' && input.isSigner !== false) {
                return render('context.identity');
            }
            return render('context.identity.publicKey');
        case 'payerValueNode':
            interfaces.add('payer');
            if (input.kind === 'instructionAccountNode' && input.isSigner !== false) {
                return render('context.payer');
            }
            return render('context.payer.publicKey');
        case 'accountBumpValueNode':
            imports.add('shared', 'expectPda');
            return render(`expectPda(resolvedAccounts.${(0, shared_1.camelCase)(defaultValue.name)}.value)[1]`);
        case 'argumentValueNode':
            imports.add('shared', 'expectSome');
            return render(`expectSome(${argObject}.${(0, shared_1.camelCase)(defaultValue.name)})`);
        case 'resolverValueNode':
            const resolverName = (0, shared_1.camelCase)(defaultValue.name);
            const isWritable = input.kind === 'instructionAccountNode' && input.isWritable
                ? 'true'
                : 'false';
            imports.add((_c = defaultValue.importFrom) !== null && _c !== void 0 ? _c : 'hooked', resolverName);
            interfaces.add(['eddsa', 'identity', 'payer']);
            return render(`${resolverName}(context, resolvedAccounts, ${argObject}, programId, ${isWritable})`);
        case 'conditionalValueNode':
            const ifTrueRenderer = renderNestedInstructionDefault(input, valueNodeVisitor, optionalAccountStrategy, defaultValue.ifTrue, argObject);
            const ifFalseRenderer = renderNestedInstructionDefault(input, valueNodeVisitor, optionalAccountStrategy, defaultValue.ifFalse, argObject);
            if (!ifTrueRenderer && !ifFalseRenderer) {
                return { imports, interfaces, render: '' };
            }
            if (ifTrueRenderer) {
                imports.mergeWith(ifTrueRenderer.imports);
                interfaces.mergeWith(ifTrueRenderer.interfaces);
            }
            if (ifFalseRenderer) {
                imports.mergeWith(ifFalseRenderer.imports);
                interfaces.mergeWith(ifFalseRenderer.interfaces);
            }
            const negatedCondition = !ifTrueRenderer;
            let condition = 'true';
            if ((0, nodes_1.isNode)(defaultValue.condition, 'resolverValueNode')) {
                const conditionalResolverName = (0, shared_1.camelCase)(defaultValue.condition.name);
                const conditionalIsWritable = input.kind === 'instructionAccountNode' && input.isWritable
                    ? 'true'
                    : 'false';
                imports.add((_d = defaultValue.condition.importFrom) !== null && _d !== void 0 ? _d : 'hooked', conditionalResolverName);
                interfaces.add(['eddsa', 'identity', 'payer']);
                condition = `${conditionalResolverName}(context, resolvedAccounts, ${argObject}, programId, ${conditionalIsWritable})`;
                condition = negatedCondition ? `!${condition}` : condition;
            }
            else {
                const comparedInputName = (0, nodes_1.isNode)(defaultValue.condition, 'accountValueNode')
                    ? `resolvedAccounts.${(0, shared_1.camelCase)(defaultValue.condition.name)}.value`
                    : `${argObject}.${(0, shared_1.camelCase)(defaultValue.condition.name)}`;
                if (defaultValue.value) {
                    const comparedValue = (0, visitors_1.visit)(defaultValue.value, valueNodeVisitor);
                    imports.mergeWith(comparedValue.imports);
                    const operator = negatedCondition ? '!==' : '===';
                    condition = `${comparedInputName} ${operator} ${comparedValue.render}`;
                }
                else {
                    condition = negatedCondition
                        ? `!${comparedInputName}`
                        : comparedInputName;
                }
            }
            if (ifTrueRenderer && ifFalseRenderer) {
                return {
                    imports,
                    interfaces,
                    render: `if (${condition}) {\n${ifTrueRenderer.render}\n} else {\n${ifFalseRenderer.render}\n}`,
                };
            }
            return {
                imports,
                interfaces,
                render: `if (${condition}) {\n${ifTrueRenderer ? ifTrueRenderer.render : ifFalseRenderer === null || ifFalseRenderer === void 0 ? void 0 : ifFalseRenderer.render}\n}`,
            };
        default:
            const valueManifest = (0, visitors_1.visit)(defaultValue, valueNodeVisitor);
            imports.mergeWith(valueManifest.imports);
            return render(valueManifest.render);
    }
}
exports.renderInstructionDefaults = renderInstructionDefaults;
function renderNestedInstructionDefault(input, valueNodeVisitor, optionalAccountStrategy, defaultValue, argObject) {
    if (!defaultValue)
        return undefined;
    return renderInstructionDefaults({ ...input, defaultValue }, valueNodeVisitor, optionalAccountStrategy, argObject);
}
//# sourceMappingURL=renderInstructionDefaults.js.map