"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.renderValueNodeVisitor = void 0;
const nodes_1 = require("../../nodes");
const shared_1 = require("../../shared");
const visitors_1 = require("../../visitors");
const JavaScriptImportMap_1 = require("./JavaScriptImportMap");
function renderValueNodeVisitor(input) {
    const { linkables, nonScalarEnums } = input;
    return {
        visitArrayValue(node) {
            const list = node.items.map((v) => (0, visitors_1.visit)(v, this));
            return {
                imports: new JavaScriptImportMap_1.JavaScriptImportMap().mergeWith(...list.map((c) => c.imports)),
                render: `[${list.map((c) => c.render).join(', ')}]`,
            };
        },
        visitBooleanValue(node) {
            return {
                imports: new JavaScriptImportMap_1.JavaScriptImportMap(),
                render: JSON.stringify(node.boolean),
            };
        },
        visitEnumValue(node) {
            var _a, _b;
            const imports = new JavaScriptImportMap_1.JavaScriptImportMap();
            const enumName = (0, shared_1.pascalCase)(node.enum.name);
            const variantName = (0, shared_1.pascalCase)(node.variant);
            const importFrom = (_a = node.enum.importFrom) !== null && _a !== void 0 ? _a : 'generatedTypes';
            const enumNode = (_b = linkables.get(node.enum)) === null || _b === void 0 ? void 0 : _b.type;
            const isScalar = enumNode && (0, nodes_1.isNode)(enumNode, 'enumTypeNode')
                ? (0, nodes_1.isScalarEnum)(enumNode)
                : !nonScalarEnums.includes(node.enum.name);
            if (!node.value && isScalar) {
                return {
                    imports: imports.add(importFrom, enumName),
                    render: `${enumName}.${variantName}`,
                };
            }
            const enumFn = (0, shared_1.camelCase)(node.enum.name);
            imports.add(importFrom, enumFn);
            if (!node.value) {
                return { imports, render: `${enumFn}('${variantName}')` };
            }
            const enumValue = (0, visitors_1.visit)(node.value, this);
            const fields = enumValue.render;
            imports.mergeWith(enumValue.imports);
            return {
                imports,
                render: `${enumFn}('${variantName}', ${fields})`,
            };
        },
        visitMapValue(node) {
            const map = node.entries.map((entry) => (0, visitors_1.visit)(entry, this));
            return {
                imports: new JavaScriptImportMap_1.JavaScriptImportMap().mergeWith(...map.map((c) => c.imports)),
                render: `new Map([${map.map((c) => c.render).join(', ')}])`,
            };
        },
        visitMapEntryValue(node) {
            const mapKey = (0, visitors_1.visit)(node.key, this);
            const mapValue = (0, visitors_1.visit)(node.value, this);
            return {
                imports: mapKey.imports.mergeWith(mapValue.imports),
                render: `[${mapKey.render}, ${mapValue.render}]`,
            };
        },
        visitNoneValue() {
            return {
                imports: new JavaScriptImportMap_1.JavaScriptImportMap().add('umi', 'none'),
                render: 'none()',
            };
        },
        visitNumberValue(node) {
            return {
                imports: new JavaScriptImportMap_1.JavaScriptImportMap(),
                render: JSON.stringify(node.number),
            };
        },
        visitPublicKeyValue(node) {
            return {
                imports: new JavaScriptImportMap_1.JavaScriptImportMap().add('umi', 'publicKey'),
                render: `publicKey("${node.publicKey}")`,
            };
        },
        visitSetValue(node) {
            const set = node.items.map((v) => (0, visitors_1.visit)(v, this));
            return {
                imports: new JavaScriptImportMap_1.JavaScriptImportMap().mergeWith(...set.map((c) => c.imports)),
                render: `new Set([${set.map((c) => c.render).join(', ')}])`,
            };
        },
        visitSomeValue(node) {
            const child = (0, visitors_1.visit)(node.value, this);
            return {
                imports: child.imports.add('umi', 'some'),
                render: `some(${child.render})`,
            };
        },
        visitStringValue(node) {
            return {
                imports: new JavaScriptImportMap_1.JavaScriptImportMap(),
                render: JSON.stringify(node.string),
            };
        },
        visitStructValue(node) {
            const struct = node.fields.map((field) => (0, visitors_1.visit)(field, this));
            return {
                imports: new JavaScriptImportMap_1.JavaScriptImportMap().mergeWith(...struct.map((c) => c.imports)),
                render: `{ ${struct.map((c) => c.render).join(', ')} }`,
            };
        },
        visitStructFieldValue(node) {
            const structValue = (0, visitors_1.visit)(node.value, this);
            return {
                imports: structValue.imports,
                render: `${node.name}: ${structValue.render}`,
            };
        },
        visitTupleValue(node) {
            const list = node.items.map((v) => (0, visitors_1.visit)(v, this));
            return {
                imports: new JavaScriptImportMap_1.JavaScriptImportMap().mergeWith(...list.map((c) => c.imports)),
                render: `[${list.map((c) => c.render).join(', ')}]`,
            };
        },
    };
}
exports.renderValueNodeVisitor = renderValueNodeVisitor;
//# sourceMappingURL=renderValueNodeVisitor.js.map