"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getRenderMapVisitor = void 0;
const nodes_1 = require("../../nodes");
const shared_1 = require("../../shared");
const visitors_1 = require("../../visitors");
const RustImportMap_1 = require("./RustImportMap");
const getTypeManifestVisitor_1 = require("./getTypeManifestVisitor");
const renderValueNodeVisitor_1 = require("./renderValueNodeVisitor");
function getRenderMapVisitor(options = {}) {
    var _a, _b;
    const linkables = new shared_1.LinkableDictionary();
    let program = null;
    const renderParentInstructions = (_a = options.renderParentInstructions) !== null && _a !== void 0 ? _a : false;
    const dependencyMap = (_b = options.dependencyMap) !== null && _b !== void 0 ? _b : {};
    const typeManifestVisitor = (0, getTypeManifestVisitor_1.getTypeManifestVisitor)();
    return (0, shared_1.pipe)((0, visitors_1.staticVisitor)(() => new shared_1.RenderMap(), [
        'rootNode',
        'programNode',
        'instructionNode',
        'accountNode',
        'definedTypeNode',
    ]), (v) => (0, visitors_1.extendVisitor)(v, {
        visitRoot(node, { self }) {
            const programsToExport = node.programs;
            const accountsToExport = (0, nodes_1.getAllAccounts)(node);
            const instructionsToExport = (0, nodes_1.getAllInstructionsWithSubs)(node, {
                leavesOnly: !renderParentInstructions,
            });
            const definedTypesToExport = (0, nodes_1.getAllDefinedTypes)(node);
            const hasAnythingToExport = programsToExport.length > 0 ||
                accountsToExport.length > 0 ||
                instructionsToExport.length > 0 ||
                definedTypesToExport.length > 0;
            const ctx = {
                root: node,
                programsToExport,
                accountsToExport,
                instructionsToExport,
                definedTypesToExport,
                hasAnythingToExport,
            };
            const map = new shared_1.RenderMap();
            if (programsToExport.length > 0) {
                map
                    .add('programs.rs', render('programsMod.njk', ctx))
                    .add('errors/mod.rs', render('errorsMod.njk', ctx));
            }
            if (accountsToExport.length > 0) {
                map.add('accounts/mod.rs', render('accountsMod.njk', ctx));
            }
            if (instructionsToExport.length > 0) {
                map.add('instructions/mod.rs', render('instructionsMod.njk', ctx));
            }
            if (definedTypesToExport.length > 0) {
                map.add('types/mod.rs', render('definedTypesMod.njk', ctx));
            }
            return map
                .add('mod.rs', render('rootMod.njk', ctx))
                .mergeWith(...node.programs.map((p) => (0, visitors_1.visit)(p, self)));
        },
        visitProgram(node, { self }) {
            program = node;
            const renderMap = new shared_1.RenderMap()
                .mergeWith(...node.accounts.map((account) => (0, visitors_1.visit)(account, self)))
                .mergeWith(...node.definedTypes.map((type) => (0, visitors_1.visit)(type, self)))
                .mergeWith(...(0, nodes_1.getAllInstructionsWithSubs)(node, {
                leavesOnly: !renderParentInstructions,
            }).map((ix) => (0, visitors_1.visit)(ix, self)));
            // Errors.
            if (node.errors.length > 0) {
                renderMap.add(`errors/${(0, shared_1.snakeCase)(node.name)}.rs`, render('errorsPage.njk', {
                    imports: new RustImportMap_1.RustImportMap().toString(dependencyMap),
                    program: node,
                    errors: node.errors.map((error) => ({
                        ...error,
                        prefixedName: (0, shared_1.pascalCase)(node.prefix) + (0, shared_1.pascalCase)(error.name),
                    })),
                }));
            }
            program = null;
            return renderMap;
        },
        visitAccount(node) {
            var _a;
            const typeManifest = (0, visitors_1.visit)(node, typeManifestVisitor);
            // Seeds.
            const seedsImports = new RustImportMap_1.RustImportMap();
            const pda = node.pda ? linkables.get(node.pda) : undefined;
            const pdaSeeds = (_a = pda === null || pda === void 0 ? void 0 : pda.seeds) !== null && _a !== void 0 ? _a : [];
            const seeds = pdaSeeds.map((seed) => {
                if ((0, nodes_1.isNode)(seed, 'constantPdaSeedNode')) {
                    const seedManifest = (0, visitors_1.visit)(seed.type, typeManifestVisitor);
                    const seedValue = seed.value;
                    const valueManifest = (0, renderValueNodeVisitor_1.renderValueNode)(seedValue, true);
                    seedValue.render = valueManifest.render;
                    seedsImports.mergeWith(valueManifest.imports);
                    return { ...seed, typeManifest: seedManifest };
                }
                if ((0, nodes_1.isNode)(seed, 'variablePdaSeedNode')) {
                    const seedManifest = (0, visitors_1.visit)(seed.type, typeManifestVisitor);
                    seedsImports.mergeWith(seedManifest.imports);
                    return { ...seed, typeManifest: seedManifest };
                }
                return seed;
            });
            const hasVariableSeeds = pdaSeeds.filter((0, nodes_1.isNodeFilter)('variablePdaSeedNode')).length > 0;
            const constantSeeds = pdaSeeds.filter((0, nodes_1.isNodeFilter)('constantPdaSeedNode'));
            const { imports } = typeManifest;
            if (hasVariableSeeds) {
                imports.mergeWith(seedsImports);
            }
            return new shared_1.RenderMap().add(`accounts/${(0, shared_1.snakeCase)(node.name)}.rs`, render('accountsPage.njk', {
                account: node,
                imports: imports
                    .remove(`generatedAccounts::${(0, shared_1.pascalCase)(node.name)}`)
                    .toString(dependencyMap),
                program,
                typeManifest,
                seeds,
                constantSeeds,
                hasVariableSeeds,
                pda,
            }));
        },
        visitInstruction(node) {
            // Imports.
            const imports = new RustImportMap_1.RustImportMap();
            imports.add(['borsh::BorshDeserialize', 'borsh::BorshSerialize']);
            // canMergeAccountsAndArgs
            const accountsAndArgsConflicts = getConflictsForInstructionAccountsAndArgs(node);
            if (accountsAndArgsConflicts.length > 0) {
                (0, shared_1.logWarn)(`[Rust] Accounts and args of instruction [${node.name}] have the following ` +
                    `conflicting attributes [${accountsAndArgsConflicts.join(', ')}]. ` +
                    `Thus, the conflicting arguments will be suffixed with "_arg". ` +
                    'You may want to rename the conflicting attributes.');
            }
            // Instruction args.
            const instructionArgs = [];
            let hasArgs = false;
            let hasOptional = false;
            node.arguments.forEach((argument) => {
                typeManifestVisitor.setParentName(`${(0, shared_1.pascalCase)(node.name)}InstructionData${(0, shared_1.pascalCase)(argument.name)}`);
                typeManifestVisitor.setNestedStruct(true);
                const manifest = (0, visitors_1.visit)(argument.type, typeManifestVisitor);
                imports.mergeWith(manifest.imports);
                const innerOptionType = (0, nodes_1.isNode)(argument.type, 'optionTypeNode')
                    ? manifest.type.slice('Option<'.length, -1)
                    : null;
                typeManifestVisitor.setParentName(null);
                typeManifestVisitor.setNestedStruct(false);
                const hasDefaultValue = !!argument.defaultValue &&
                    (0, nodes_1.isNode)(argument.defaultValue, nodes_1.VALUE_NODES);
                let renderValue = null;
                if (hasDefaultValue) {
                    const { imports: argImports, render: value } = (0, renderValueNodeVisitor_1.renderValueNode)(argument.defaultValue);
                    imports.mergeWith(argImports);
                    renderValue = value;
                }
                hasArgs = hasArgs || argument.defaultValueStrategy !== 'omitted';
                hasOptional =
                    hasOptional ||
                        (hasDefaultValue && argument.defaultValueStrategy !== 'omitted');
                const name = accountsAndArgsConflicts.includes(argument.name)
                    ? `${argument.name}_arg`
                    : argument.name;
                instructionArgs.push({
                    name,
                    type: manifest.type,
                    default: hasDefaultValue && argument.defaultValueStrategy === 'omitted',
                    optional: hasDefaultValue && argument.defaultValueStrategy !== 'omitted',
                    innerOptionType,
                    value: renderValue,
                });
            });
            const struct = (0, nodes_1.structTypeNodeFromInstructionArgumentNodes)(node.arguments);
            typeManifestVisitor.setParentName(`${(0, shared_1.pascalCase)(node.name)}InstructionData`);
            const typeManifest = (0, visitors_1.visit)(struct, typeManifestVisitor);
            typeManifestVisitor.setParentName(null);
            return new shared_1.RenderMap().add(`instructions/${(0, shared_1.snakeCase)(node.name)}.rs`, render('instructionsPage.njk', {
                instruction: node,
                imports: imports
                    .remove(`generatedInstructions::${(0, shared_1.pascalCase)(node.name)}`)
                    .toString(dependencyMap),
                instructionArgs,
                hasArgs,
                hasOptional,
                program,
                typeManifest,
            }));
        },
        visitDefinedType(node) {
            const typeManifest = (0, visitors_1.visit)(node, typeManifestVisitor);
            const imports = new RustImportMap_1.RustImportMap().mergeWithManifest(typeManifest);
            return new shared_1.RenderMap().add(`types/${(0, shared_1.snakeCase)(node.name)}.rs`, render('definedTypesPage.njk', {
                definedType: node,
                imports: imports
                    .remove(`generatedTypes::${(0, shared_1.pascalCase)(node.name)}`)
                    .toString(dependencyMap),
                typeManifest,
            }));
        },
    }), (v) => (0, visitors_1.recordLinkablesVisitor)(v, linkables));
}
exports.getRenderMapVisitor = getRenderMapVisitor;
function render(template, context, renderOptions) {
    return (0, shared_1.resolveTemplate)(`${__dirname}/templates`, template, context, renderOptions);
}
function getConflictsForInstructionAccountsAndArgs(instruction) {
    const allNames = [
        ...instruction.accounts.map((account) => account.name),
        ...instruction.arguments.map((argument) => argument.name),
    ];
    const duplicates = allNames.filter((e, i, a) => a.indexOf(e) !== i);
    return [...new Set(duplicates)];
}
//# sourceMappingURL=getRenderMapVisitor.js.map