"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.fillDefaultPdaSeedValuesVisitor = void 0;
const nodes_1 = require("../nodes");
const shared_1 = require("../shared");
const extendVisitor_1 = require("./extendVisitor");
const identityVisitor_1 = require("./identityVisitor");
/**
 * Fills in default values for variable PDA seeds that are not explicitly provided.
 * Namely, public key seeds are filled with an accountValueNode using the seed name
 * and other types of seeds are filled with an argumentValueNode using the seed name.
 *
 * An instruction and linkable dictionary are required to determine which seeds are
 * valids and to find the pdaLinkNode for the seed respectively. Any invalid default
 * seed won't be filled in.
 *
 * Strict mode goes one step further and will throw an error if the final array of
 * pdaSeedValueNodes contains invalid seeds or if there aren't enough variable seeds.
 */
function fillDefaultPdaSeedValuesVisitor(instruction, linkables, strictMode = false) {
    return (0, shared_1.pipe)((0, identityVisitor_1.identityVisitor)(nodes_1.INSTRUCTION_INPUT_VALUE_NODE), (v) => (0, extendVisitor_1.extendVisitor)(v, {
        visitPdaValue(node, { next }) {
            const visitedNode = next(node);
            (0, nodes_1.assertIsNode)(visitedNode, 'pdaValueNode');
            const foundPda = linkables.get(visitedNode.pda);
            if (!foundPda)
                return visitedNode;
            const seeds = addDefaultSeedValuesFromPdaWhenMissing(instruction, foundPda, visitedNode.seeds);
            if (strictMode && !allSeedsAreValid(instruction, foundPda, seeds)) {
                throw new shared_1.KinobiError(`Invalid seed values for PDA ${foundPda.name} in instruction ${instruction.name}`);
            }
            return (0, nodes_1.pdaValueNode)(visitedNode.pda, seeds);
        },
    }));
}
exports.fillDefaultPdaSeedValuesVisitor = fillDefaultPdaSeedValuesVisitor;
function addDefaultSeedValuesFromPdaWhenMissing(instruction, pda, existingSeeds) {
    const existingSeedNames = new Set(existingSeeds.map((seed) => seed.name));
    const defaultSeeds = getDefaultSeedValuesFromPda(instruction, pda).filter((seed) => !existingSeedNames.has(seed.name));
    return [...existingSeeds, ...defaultSeeds];
}
function getDefaultSeedValuesFromPda(instruction, pda) {
    return pda.seeds.flatMap((seed) => {
        if (!(0, nodes_1.isNode)(seed, 'variablePdaSeedNode'))
            return [];
        const hasMatchingAccount = instruction.accounts.some((a) => a.name === seed.name);
        if ((0, nodes_1.isNode)(seed.type, 'publicKeyTypeNode') && hasMatchingAccount) {
            return [(0, nodes_1.pdaSeedValueNode)(seed.name, (0, nodes_1.accountValueNode)(seed.name))];
        }
        const hasMatchingArgument = (0, nodes_1.getAllInstructionArguments)(instruction).some((a) => a.name === seed.name);
        if (hasMatchingArgument) {
            return [(0, nodes_1.pdaSeedValueNode)(seed.name, (0, nodes_1.argumentValueNode)(seed.name))];
        }
        return [];
    });
}
function allSeedsAreValid(instruction, foundPda, seeds) {
    const hasAllVariableSeeds = foundPda.seeds.filter((0, nodes_1.isNodeFilter)('variablePdaSeedNode')).length ===
        seeds.length;
    const allAccountsName = instruction.accounts.map((a) => a.name);
    const allArgumentsName = (0, nodes_1.getAllInstructionArguments)(instruction).map((a) => a.name);
    const validSeeds = seeds.every((seed) => {
        if ((0, nodes_1.isNode)(seed.value, 'accountValueNode')) {
            return allAccountsName.includes(seed.value.name);
        }
        if ((0, nodes_1.isNode)(seed.value, 'argumentValueNode')) {
            return allArgumentsName.includes(seed.value.name);
        }
        return true;
    });
    return hasAllVariableSeeds && validSeeds;
}
//# sourceMappingURL=fillDefaultPdaSeedValuesVisitor.js.map