"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getByteSizeVisitor = void 0;
const nodes_1 = require("../nodes");
const mergeVisitor_1 = require("./mergeVisitor");
const visitor_1 = require("./visitor");
function getByteSizeVisitor(linkables) {
    const visitedDefinedTypes = new Map();
    const definedTypeStack = [];
    const sumSizes = (values) => values.reduce((all, one) => (all === null || one === null ? null : all + one), 0);
    const visitor = (0, mergeVisitor_1.mergeVisitor)(() => null, (_, values) => sumSizes(values), [
        ...nodes_1.REGISTERED_TYPE_NODE_KINDS,
        'definedTypeLinkNode',
        'definedTypeNode',
        'accountNode',
        'instructionNode',
        'instructionArgumentNode',
    ]);
    return {
        ...visitor,
        visitAccount(node) {
            return (0, visitor_1.visit)(node.data, this);
        },
        visitInstruction(node) {
            return sumSizes(node.arguments.map((arg) => (0, visitor_1.visit)(arg, this)));
        },
        visitInstructionArgument(node) {
            return (0, visitor_1.visit)(node.type, this);
        },
        visitDefinedType(node) {
            if (visitedDefinedTypes.has(node.name)) {
                return visitedDefinedTypes.get(node.name);
            }
            definedTypeStack.push(node.name);
            const child = (0, visitor_1.visit)(node.type, this);
            definedTypeStack.pop();
            visitedDefinedTypes.set(node.name, child);
            return child;
        },
        visitArrayType(node) {
            if (!(0, nodes_1.isNode)(node.size, 'fixedSizeNode'))
                return null;
            const fixedSize = node.size.size;
            const itemSize = (0, visitor_1.visit)(node.item, this);
            const arraySize = itemSize !== null ? itemSize * fixedSize : null;
            return fixedSize === 0 ? 0 : arraySize;
        },
        visitDefinedTypeLink(node) {
            if (node.importFrom)
                return null;
            // Fetch the linked type and return null if not found.
            // The validator visitor will throw a proper error later on.
            const linkedDefinedType = linkables.get(node);
            if (!linkedDefinedType) {
                return null;
            }
            // This prevents infinite recursion by using assuming
            // cyclic types don't have a fixed size.
            if (definedTypeStack.includes(linkedDefinedType.name)) {
                return null;
            }
            return (0, visitor_1.visit)(linkedDefinedType, this);
        },
        visitEnumType(node) {
            var _a;
            const prefix = (_a = (0, visitor_1.visit)(node.size, this)) !== null && _a !== void 0 ? _a : 1;
            if ((0, nodes_1.isScalarEnum)(node))
                return prefix;
            const variantSizes = node.variants.map((v) => (0, visitor_1.visit)(v, this));
            const allVariantHaveTheSameFixedSize = variantSizes.every((one, i, all) => one === all[0]);
            return allVariantHaveTheSameFixedSize &&
                variantSizes.length > 0 &&
                variantSizes[0] !== null
                ? variantSizes[0] + prefix
                : null;
        },
        visitEnumEmptyVariantType() {
            return 0;
        },
        visitOptionType(node) {
            if (!node.fixed)
                return null;
            const prefixSize = (0, visitor_1.visit)(node.prefix, this);
            const itemSize = (0, visitor_1.visit)(node.item, this);
            return itemSize !== null ? itemSize + prefixSize : null;
        },
        visitBytesType(node) {
            if (!(0, nodes_1.isNode)(node.size, 'fixedSizeNode'))
                return null;
            return node.size.size;
        },
        visitNumberType(node) {
            return parseInt(node.format.slice(1), 10) / 8;
        },
        visitPublicKeyType() {
            return 32;
        },
        visitStringType(node) {
            if (!(0, nodes_1.isNode)(node.size, 'fixedSizeNode'))
                return null;
            return node.size.size;
        },
    };
}
exports.getByteSizeVisitor = getByteSizeVisitor;
//# sourceMappingURL=getByteSizeVisitor.js.map