"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.mergeVisitor = void 0;
const nodes_1 = require("../nodes");
const staticVisitor_1 = require("./staticVisitor");
const visitor_1 = require("./visitor");
function mergeVisitor(leafValue, merge, nodeKeys = nodes_1.REGISTERED_NODE_KINDS) {
    const castedNodeKeys = nodeKeys;
    const visitor = (0, staticVisitor_1.staticVisitor)(leafValue, castedNodeKeys);
    const visit = (v) => (node) => castedNodeKeys.includes(node.kind) ? [(0, visitor_1.visit)(node, v)] : [];
    if (castedNodeKeys.includes('rootNode')) {
        visitor.visitRoot = function visitRoot(node) {
            return merge(node, node.programs.flatMap(visit(this)));
        };
    }
    if (castedNodeKeys.includes('programNode')) {
        visitor.visitProgram = function visitProgram(node) {
            return merge(node, [
                ...node.pdas.flatMap(visit(this)),
                ...node.accounts.flatMap(visit(this)),
                ...node.instructions.flatMap(visit(this)),
                ...node.definedTypes.flatMap(visit(this)),
                ...node.errors.flatMap(visit(this)),
            ]);
        };
    }
    if (castedNodeKeys.includes('pdaNode')) {
        visitor.visitPda = function visitPda(node) {
            return merge(node, node.seeds.flatMap(visit(this)));
        };
    }
    if (castedNodeKeys.includes('accountNode')) {
        visitor.visitAccount = function visitAccount(node) {
            var _a;
            return merge(node, [
                ...visit(this)(node.data),
                ...(node.pda ? visit(this)(node.pda) : []),
                ...((_a = node.discriminators) !== null && _a !== void 0 ? _a : []).flatMap(visit(this)),
            ]);
        };
    }
    if (castedNodeKeys.includes('instructionNode')) {
        visitor.visitInstruction = function visitInstruction(node) {
            var _a, _b, _c, _d, _e;
            return merge(node, [
                ...node.accounts.flatMap(visit(this)),
                ...node.arguments.flatMap(visit(this)),
                ...((_a = node.extraArguments) !== null && _a !== void 0 ? _a : []).flatMap(visit(this)),
                ...((_b = node.remainingAccounts) !== null && _b !== void 0 ? _b : []).flatMap(visit(this)),
                ...((_c = node.byteDeltas) !== null && _c !== void 0 ? _c : []).flatMap(visit(this)),
                ...((_d = node.discriminators) !== null && _d !== void 0 ? _d : []).flatMap(visit(this)),
                ...((_e = node.subInstructions) !== null && _e !== void 0 ? _e : []).flatMap(visit(this)),
            ]);
        };
    }
    if (castedNodeKeys.includes('instructionAccountNode')) {
        visitor.visitInstructionAccount = function visitInstructionAccount(node) {
            return merge(node, [
                ...(node.defaultValue ? visit(this)(node.defaultValue) : []),
            ]);
        };
    }
    if (castedNodeKeys.includes('instructionArgumentNode')) {
        visitor.visitInstructionArgument = function visitInstructionArgument(node) {
            return merge(node, [
                ...visit(this)(node.type),
                ...(node.defaultValue ? visit(this)(node.defaultValue) : []),
            ]);
        };
    }
    if (castedNodeKeys.includes('instructionRemainingAccountsNode')) {
        visitor.visitInstructionRemainingAccounts =
            function visitInstructionRemainingAccounts(node) {
                return merge(node, visit(this)(node.value));
            };
    }
    if (castedNodeKeys.includes('instructionByteDeltaNode')) {
        visitor.visitInstructionByteDelta = function visitInstructionByteDelta(node) {
            return merge(node, visit(this)(node.value));
        };
    }
    if (castedNodeKeys.includes('definedTypeNode')) {
        visitor.visitDefinedType = function visitDefinedType(node) {
            return merge(node, visit(this)(node.type));
        };
    }
    if (castedNodeKeys.includes('arrayTypeNode')) {
        visitor.visitArrayType = function visitArrayType(node) {
            return merge(node, [
                ...visit(this)(node.size),
                ...visit(this)(node.item),
            ]);
        };
    }
    if (castedNodeKeys.includes('enumTypeNode')) {
        visitor.visitEnumType = function visitEnumType(node) {
            return merge(node, [
                ...visit(this)(node.size),
                ...node.variants.flatMap(visit(this)),
            ]);
        };
    }
    if (castedNodeKeys.includes('enumStructVariantTypeNode')) {
        visitor.visitEnumStructVariantType = function visitEnumStructVariantType(node) {
            return merge(node, visit(this)(node.struct));
        };
    }
    if (castedNodeKeys.includes('enumTupleVariantTypeNode')) {
        visitor.visitEnumTupleVariantType = function visitEnumTupleVariantType(node) {
            return merge(node, visit(this)(node.tuple));
        };
    }
    if (castedNodeKeys.includes('mapTypeNode')) {
        visitor.visitMapType = function visitMapType(node) {
            return merge(node, [
                ...visit(this)(node.size),
                ...visit(this)(node.key),
                ...visit(this)(node.value),
            ]);
        };
    }
    if (castedNodeKeys.includes('optionTypeNode')) {
        visitor.visitOptionType = function visitOptionType(node) {
            return merge(node, [
                ...visit(this)(node.prefix),
                ...visit(this)(node.item),
            ]);
        };
    }
    if (castedNodeKeys.includes('booleanTypeNode')) {
        visitor.visitBooleanType = function visitBooleanType(node) {
            return merge(node, visit(this)(node.size));
        };
    }
    if (castedNodeKeys.includes('setTypeNode')) {
        visitor.visitSetType = function visitSetType(node) {
            return merge(node, [
                ...visit(this)(node.size),
                ...visit(this)(node.item),
            ]);
        };
    }
    if (castedNodeKeys.includes('structTypeNode')) {
        visitor.visitStructType = function visitStructType(node) {
            return merge(node, node.fields.flatMap(visit(this)));
        };
    }
    if (castedNodeKeys.includes('structFieldTypeNode')) {
        visitor.visitStructFieldType = function visitStructFieldType(node) {
            return merge(node, [
                ...visit(this)(node.type),
                ...(node.defaultValue ? visit(this)(node.defaultValue) : []),
            ]);
        };
    }
    if (castedNodeKeys.includes('tupleTypeNode')) {
        visitor.visitTupleType = function visitTupleType(node) {
            return merge(node, node.items.flatMap(visit(this)));
        };
    }
    if (castedNodeKeys.includes('amountTypeNode')) {
        visitor.visitAmountType = function visitAmountType(node) {
            return merge(node, visit(this)(node.number));
        };
    }
    if (castedNodeKeys.includes('dateTimeTypeNode')) {
        visitor.visitDateTimeType = function visitDateTimeType(node) {
            return merge(node, visit(this)(node.number));
        };
    }
    if (castedNodeKeys.includes('solAmountTypeNode')) {
        visitor.visitSolAmountType = function visitSolAmountType(node) {
            return merge(node, visit(this)(node.number));
        };
    }
    if (castedNodeKeys.includes('stringTypeNode')) {
        visitor.visitStringType = function visitStringType(node) {
            return merge(node, visit(this)(node.size));
        };
    }
    if (castedNodeKeys.includes('bytesTypeNode')) {
        visitor.visitBytesType = function visitBytesType(node) {
            return merge(node, visit(this)(node.size));
        };
    }
    if (castedNodeKeys.includes('prefixedSizeNode')) {
        visitor.visitPrefixedSize = function visitPrefixedSize(node) {
            return merge(node, visit(this)(node.prefix));
        };
    }
    if (castedNodeKeys.includes('arrayValueNode')) {
        visitor.visitArrayValue = function visitArrayValue(node) {
            return merge(node, node.items.flatMap(visit(this)));
        };
    }
    if (castedNodeKeys.includes('enumValueNode')) {
        visitor.visitEnumValue = function visitEnumValue(node) {
            return merge(node, [
                ...visit(this)(node.enum),
                ...(node.value ? visit(this)(node.value) : []),
            ]);
        };
    }
    if (castedNodeKeys.includes('mapValueNode')) {
        visitor.visitMapValue = function visitMapValue(node) {
            return merge(node, node.entries.flatMap(visit(this)));
        };
    }
    if (castedNodeKeys.includes('mapEntryValueNode')) {
        visitor.visitMapEntryValue = function visitMapEntryValue(node) {
            return merge(node, [
                ...visit(this)(node.key),
                ...visit(this)(node.value),
            ]);
        };
    }
    if (castedNodeKeys.includes('setValueNode')) {
        visitor.visitSetValue = function visitSetValue(node) {
            return merge(node, node.items.flatMap(visit(this)));
        };
    }
    if (castedNodeKeys.includes('someValueNode')) {
        visitor.visitSomeValue = function visitSomeValue(node) {
            return merge(node, visit(this)(node.value));
        };
    }
    if (castedNodeKeys.includes('structValueNode')) {
        visitor.visitStructValue = function visitStructValue(node) {
            return merge(node, node.fields.flatMap(visit(this)));
        };
    }
    if (castedNodeKeys.includes('structFieldValueNode')) {
        visitor.visitStructFieldValue = function visitStructFieldValue(node) {
            return merge(node, visit(this)(node.value));
        };
    }
    if (castedNodeKeys.includes('tupleValueNode')) {
        visitor.visitTupleValue = function visitTupleValue(node) {
            return merge(node, node.items.flatMap(visit(this)));
        };
    }
    if (castedNodeKeys.includes('constantPdaSeedNode')) {
        visitor.visitConstantPdaSeed = function visitConstantPdaSeed(node) {
            return merge(node, [
                ...visit(this)(node.type),
                ...visit(this)(node.value),
            ]);
        };
    }
    if (castedNodeKeys.includes('variablePdaSeedNode')) {
        visitor.visitVariablePdaSeed = function visitVariablePdaSeed(node) {
            return merge(node, visit(this)(node.type));
        };
    }
    if (castedNodeKeys.includes('resolverValueNode')) {
        visitor.visitResolverValue = function visitResolverValue(node) {
            var _a;
            return merge(node, ((_a = node.dependsOn) !== null && _a !== void 0 ? _a : []).flatMap(visit(this)));
        };
    }
    if (castedNodeKeys.includes('conditionalValueNode')) {
        visitor.visitConditionalValue = function visitConditionalValue(node) {
            return merge(node, [
                ...visit(this)(node.condition),
                ...(node.value ? visit(this)(node.value) : []),
                ...(node.ifTrue ? visit(this)(node.ifTrue) : []),
                ...(node.ifFalse ? visit(this)(node.ifFalse) : []),
            ]);
        };
    }
    if (castedNodeKeys.includes('pdaValueNode')) {
        visitor.visitPdaValue = function visitPdaValue(node) {
            return merge(node, [
                ...visit(this)(node.pda),
                ...node.seeds.flatMap(visit(this)),
            ]);
        };
    }
    if (castedNodeKeys.includes('pdaSeedValueNode')) {
        visitor.visitPdaSeedValue = function visitPdaSeedValue(node) {
            return merge(node, visit(this)(node.value));
        };
    }
    return visitor;
}
exports.mergeVisitor = mergeVisitor;
//# sourceMappingURL=mergeVisitor.js.map