"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.setInstructionAccountDefaultValuesVisitor = exports.DEFAULT_INSTRUCTION_ACCOUNT_DEFAULT_RULES = void 0;
const nodes_1 = require("../nodes");
const contextualValueNodes_1 = require("../nodes/contextualValueNodes");
const valueNodes_1 = require("../nodes/valueNodes");
const shared_1 = require("../shared");
const extendVisitor_1 = require("./extendVisitor");
const fillDefaultPdaSeedValuesVisitor_1 = require("./fillDefaultPdaSeedValuesVisitor");
const nonNullableIdentityVisitor_1 = require("./nonNullableIdentityVisitor");
const recordLinkablesVisitor_1 = require("./recordLinkablesVisitor");
const visitor_1 = require("./visitor");
exports.DEFAULT_INSTRUCTION_ACCOUNT_DEFAULT_RULES = [
    {
        account: /^payer|feePayer$/,
        defaultValue: (0, contextualValueNodes_1.payerValueNode)(),
        ignoreIfOptional: true,
    },
    {
        account: /^authority$/,
        defaultValue: (0, contextualValueNodes_1.identityValueNode)(),
        ignoreIfOptional: true,
    },
    {
        account: /^programId$/,
        defaultValue: (0, contextualValueNodes_1.programIdValueNode)(),
        ignoreIfOptional: true,
    },
    {
        account: /^systemProgram|splSystemProgram$/,
        defaultValue: (0, valueNodes_1.publicKeyValueNode)('11111111111111111111111111111111', 'splSystem'),
        ignoreIfOptional: true,
    },
    {
        account: /^tokenProgram|splTokenProgram$/,
        defaultValue: (0, valueNodes_1.publicKeyValueNode)('TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA', 'splToken'),
        ignoreIfOptional: true,
    },
    {
        account: /^ataProgram|splAtaProgram$/,
        defaultValue: (0, valueNodes_1.publicKeyValueNode)('ATokenGPvbdGVxr1b2hvZbsiqW5xWH25efTNsLJA8knL', 'splAssociatedToken'),
        ignoreIfOptional: true,
    },
    {
        account: /^tokenMetadataProgram|mplTokenMetadataProgram$/,
        defaultValue: (0, valueNodes_1.publicKeyValueNode)('metaqbxxUerdq28cj1RbAWkYQm3ybzjb6a8bt518x1s', 'mplTokenMetadata'),
        ignoreIfOptional: true,
    },
    {
        account: /^(tokenAuth|mplTokenAuth|authorization|mplAuthorization|auth|mplAuth)RulesProgram$/,
        defaultValue: (0, valueNodes_1.publicKeyValueNode)('auth9SigNpDKz4sJJ1DfCTuZrZNSAgh9sFD3rboVmgg', 'mplTokenAuthRules'),
        ignoreIfOptional: true,
    },
    {
        account: /^candyMachineProgram|mplCandyMachineProgram$/,
        defaultValue: (0, valueNodes_1.publicKeyValueNode)('CndyV3LdqHUfDLmE5naZjVN8rBZz4tqhdefbAnjHG3JR', 'mplCandyMachine'),
        ignoreIfOptional: true,
    },
    {
        account: /^candyGuardProgram|mplCandyGuardProgram$/,
        defaultValue: (0, valueNodes_1.publicKeyValueNode)('Guard1JwRhJkVH6XZhzoYxeBVQe872VH6QggF4BWmS9g', 'mplCandyGuard'),
        ignoreIfOptional: true,
    },
    {
        account: /^clockSysvar|sysvarClock$/,
        defaultValue: (0, valueNodes_1.publicKeyValueNode)('SysvarC1ock11111111111111111111111111111111'),
        ignoreIfOptional: true,
    },
    {
        account: /^epochScheduleSysvar|sysvarEpochSchedule$/,
        defaultValue: (0, valueNodes_1.publicKeyValueNode)('SysvarEpochSchedu1e111111111111111111111111'),
        ignoreIfOptional: true,
    },
    {
        account: /^(instructions?Sysvar|sysvarInstructions?)(Account)?$/,
        defaultValue: (0, valueNodes_1.publicKeyValueNode)('Sysvar1nstructions1111111111111111111111111'),
        ignoreIfOptional: true,
    },
    {
        account: /^recentBlockhashesSysvar|sysvarRecentBlockhashes$/,
        defaultValue: (0, valueNodes_1.publicKeyValueNode)('SysvarRecentB1ockHashes11111111111111111111'),
        ignoreIfOptional: true,
    },
    {
        account: /^rent|rentSysvar|sysvarRent$/,
        defaultValue: (0, valueNodes_1.publicKeyValueNode)('SysvarRent111111111111111111111111111111111'),
        ignoreIfOptional: true,
    },
    {
        account: /^rewardsSysvar|sysvarRewards$/,
        defaultValue: (0, valueNodes_1.publicKeyValueNode)('SysvarRewards111111111111111111111111111111'),
        ignoreIfOptional: true,
    },
    {
        account: /^slotHashesSysvar|sysvarSlotHashes$/,
        defaultValue: (0, valueNodes_1.publicKeyValueNode)('SysvarS1otHashes111111111111111111111111111'),
        ignoreIfOptional: true,
    },
    {
        account: /^slotHistorySysvar|sysvarSlotHistory$/,
        defaultValue: (0, valueNodes_1.publicKeyValueNode)('SysvarS1otHistory11111111111111111111111111'),
        ignoreIfOptional: true,
    },
    {
        account: /^stakeHistorySysvar|sysvarStakeHistory$/,
        defaultValue: (0, valueNodes_1.publicKeyValueNode)('SysvarStakeHistory1111111111111111111111111'),
        ignoreIfOptional: true,
    },
    {
        account: /^mplCoreProgram$/,
        defaultValue: (0, valueNodes_1.publicKeyValueNode)('CoREENxT6tW1HoK8ypY1SxRMZTcVPm7R94rH4PZNhX7d', 'mplCore'),
        ignoreIfOptional: true,
    },
];
function setInstructionAccountDefaultValuesVisitor(rules) {
    const linkables = new shared_1.LinkableDictionary();
    // Place the rules with instructions first.
    const sortedRules = rules.sort((a, b) => {
        const ia = 'instruction' in a;
        const ib = 'instruction' in b;
        if ((ia && ib) || (!a && !ib))
            return 0;
        return ia ? -1 : 1;
    });
    function matchRule(instruction, account) {
        return sortedRules.find((rule) => {
            if ('instruction' in rule &&
                rule.instruction &&
                (0, shared_1.mainCase)(rule.instruction) !== instruction.name) {
                return false;
            }
            return typeof rule.account === 'string'
                ? (0, shared_1.mainCase)(rule.account) === account.name
                : rule.account.test(account.name);
        });
    }
    return (0, shared_1.pipe)((0, nonNullableIdentityVisitor_1.nonNullableIdentityVisitor)(['rootNode', 'programNode', 'instructionNode']), (v) => (0, recordLinkablesVisitor_1.recordLinkablesVisitor)(v, linkables), (v) => (0, extendVisitor_1.extendVisitor)(v, {
        visitInstruction(node) {
            const instructionAccounts = node.accounts.map((account) => {
                var _a;
                const rule = matchRule(node, account);
                if (!rule)
                    return account;
                if (((_a = rule.ignoreIfOptional) !== null && _a !== void 0 ? _a : false) &&
                    (account.isOptional || !!account.defaultValue)) {
                    return account;
                }
                try {
                    return {
                        ...account,
                        defaultValue: (0, visitor_1.visit)(rule.defaultValue, (0, fillDefaultPdaSeedValuesVisitor_1.fillDefaultPdaSeedValuesVisitor)(node, linkables, true)),
                    };
                }
                catch (error) {
                    return account;
                }
            });
            return (0, nodes_1.instructionNode)({
                ...node,
                accounts: instructionAccounts,
            });
        },
    }));
}
exports.setInstructionAccountDefaultValuesVisitor = setInstructionAccountDefaultValuesVisitor;
//# sourceMappingURL=setInstructionAccountDefaultValuesVisitor.js.map